﻿DeclareModule Grh

Enumeration
  #Log_Info
  #Log_Warning
  #Log_Error
EndEnumeration

Structure GraphInit
  GraphGadget.i
  ScrollGadget.i
  Font.i
  FontName.s
  FontSize.a
  *LogFunct
EndStructure

Structure GraphColor
  Back.l    ; Фоновый цвет
  Grid.l    ; Цвет линий сетки
  Text.l    ; Текст
  Sensor1.l ; Датчик 1 (ав. ТСА)
  Sensor2.l ; Датчик 2 (ав. Конденсатор)
  Sensor3.l ; Датчик 3 (д. Колонна)
  Sensor4.l ; Датчик 4 (д. Куб)
  Sensor5.l ; Датчик 5 (д. Брага)
  Sensor6.l ; Датчик № 6.
EndStructure

Structure GraphSetting
  Color.GraphColor
  NewDay.a     ; Сохранять график в файл при смене календарного дня.
  AutoScroll.a ; Автопрокрутка графика.
EndStructure

Structure GraphDataSensor
  Sensor1.w ; Датчик 1 (ав. ТСА)
  Sensor2.w ; Датчик 2 (ав. Конденсатор)
  Sensor3.w ; Датчик 3 (д. Колонна)
  Sensor4.w ; Датчик 4 (д. Куб)
  Sensor5.w ; Датчик 5 (д. Брага)
  Sensor6.w ; Датчик № 6.
EndStructure

Structure GraphData
  StructureUnion
    Sensor.GraphDataSensor
    Arr.w[6]
  EndStructureUnion
EndStructure

Declare GraphInit(*Graph.GraphInit)
Declare GraphSetting(*s.GraphSetting)
Declare AddData(*Data.GraphData, Date) ; Добавление данных для графика.
Declare GraphDraw()                    ; Перерисовка графика.
Declare Scroll_CB()

EndDeclareModule

Module Grh

UsePNGImageEncoder()

#PointSec = 5      ; Количество точек на одно измерение.
#ImgMaxWidth=30000 ; Маскимальная ширина картинки.

Prototype pLog(String.s, Type)

Structure DataGraph
  Sensors.GraphData
  Date.l
EndStructure

Structure Param Extends GraphSetting
  *LogFunct.pLog
  GraphGadget.i
  ScrollGadget.i
  FontID.i
  VisibleTime.l ; Видимая область графика в секундах.
  PointY.f      ; Количество отсчтетов на точку по координате Y.
  ScrollPos.l   ; Текущая позиция полосы прокрутки.
  MinTermo.w    ; Минимальная температура.
  MaxTermo.w    ; Максимальная температура.
EndStructure

Define ThreadID.i
Global gParam.Param
Global NewList gList.DataGraph()
EnableExplicit

;- Private procedure.

Procedure DrawVerticalLines(Date, x, w, *Area.RECT, mDate=#False)
  Protected Temp, String.s
   
  If mDate
    Temp = Second(Date)%5
    Temp=5-Temp
    Date+Temp
    x-Temp
  EndIf
  
  While x>2
    If x<w
      Line(x+*Area\left, *Area\top, 1, *Area\bottom-*Area\top+2, gParam\Color\Grid)
    EndIf
    x-5*#PointSec
  Wend
  
EndProcedure

Procedure GetTermo(*s.GraphData)
  Protected i, z, Termo.w
  
  For i=0 To 5
    *s\Arr[i] = -2000
  Next i
  
  For i=0 To 5
    Termo = gList()\Sensors\Arr[i]
    If Termo>=-550 And Termo<=1250
      If Termo<gParam\MinTermo
        Termo=gParam\MinTermo
      ElseIf Termo>gParam\MaxTermo
        Termo=gParam\MaxTermo
      EndIf
      *s\Arr[i] = Termo
    EndIf
  Next i
  
EndProcedure

Procedure DrawInfo(*Area.RECT, DrawTime=#False) ; Отрисовка графика.
  Protected Count, w, h, i, x, Color.l
  Protected ListPos.l, MaxTermo, Temp, fDate.l=0, fx, Date
  Protected s.GraphData, PosY, String.s
  Protected Dim OldPosX(5), Dim OldPosY(5)
  
  Count=ListSize(gList())
  
  *Area\top+1
  *Area\right-2
  *Area\bottom-2
  w=*Area\right-*Area\left
  h=*Area\bottom-*Area\top
  MaxTermo = gParam\MaxTermo-gParam\MinTermo
  
  For i=0 To 5
    OldPosX(i)=-1
    OldPosY(i)=-1
  Next i
  
  If Count>0 And w>0 And h>0
    
    DrawingMode(#PB_2DDrawing_Transparent)
    x=w
    
    If Count>gParam\VisibleTime
      ListPos=gParam\VisibleTime+gParam\ScrollPos-1
    Else
      ListPos=Count-1
    EndIf
    
    If ListPos>=Count
      ListPos=Count-1
    EndIf
    
    If SelectElement(gList(), ListPos)
      
      If DrawTime
        DrawText(4, 4, FormatDate("%hh:%ii:%ss", gList()\Date), gParam\Color\Text)
      EndIf
      
      While x>-#PointSec
        
        If x<0
          x=0
        EndIf
        
        GetTermo(@s)
        
        Date=gList()\Date
        
        If fDate=0
          fDate=Date
          Temp = Second(fDate)%5
          fDate+Temp
          fx=x-Temp*#PointSec
        EndIf
        
        For i=0 To 5
          
          Select i
            Case 0 : Color = gParam\Color\Sensor1
            Case 1 : Color = gParam\Color\Sensor2
            Case 2 : Color = gParam\Color\Sensor3
            Case 3 : Color = gParam\Color\Sensor4
            Case 4 : Color = gParam\Color\Sensor5
            Case 5 : Color = gParam\Color\Sensor6
          EndSelect
          
          Temp=Second(Date)
          If Temp%5=0 And x<w And x>=2
            Line(x+*Area\left, *Area\top, 1, h+2, gParam\Color\Grid)
            If Temp%10=0
              If Temp<5
                String=FormatDate("%hh:%ii", Date)
              Else
                String=FormatDate("%ss", Date)
              EndIf
              DrawText(x+*Area\left-TextWidth(String)/2, *Area\bottom+4, String, gParam\Color\Text)
            EndIf
            fDate=Date+5
            fx=x-5*#PointSec
          EndIf
          
          If s\Arr[i]>=gParam\MinTermo And s\Arr[i]<=gParam\MaxTermo
            PosY = Round((MaxTermo-(s\Arr[i]-gParam\MinTermo))/(MaxTermo/h)+*Area\top, #PB_Round_Up)
            
            If PosY>=*Area\top And PosY<=*Area\bottom
              
              If OldPosX(i)>=*Area\left And OldPosX(i)=<*Area\right And
                 OldPosY(i)>=*Area\top And OldPosY(i)<=*Area\bottom
                
                LineXY(x+*Area\left, PosY, OldPosX(i), OldPosY(i), Color)
              Else
                Plot(x+*Area\left, PosY, Color)
              EndIf
              
              OldPosX(i)=x+*Area\left
              OldPosY(i)=PosY
              
            EndIf
          Else
            OldPosX(i)=-1
            OldPosY(i)=-1
          EndIf
          
        Next i
        
        If PreviousElement(gList())=0
          Break
        EndIf
        
        x-#PointSec
      Wend
      
    EndIf
    
    If x>0 And Count<=gParam\VisibleTime And fDate<>0
      DrawVerticalLines(fDate, fx, w-4, *Area)
    EndIf
    
  ElseIf Count=0
    DrawVerticalLines(Date(), w, w-4, *Area, #True)
  EndIf
  
EndProcedure


Procedure DrawTop(w, h, DrawDate=#False) ; Рисование содержимого выше графика.
  Protected PosX, PosY, x, s
  Protected String.s, Width
  
  DrawingMode(#PB_2DDrawing_Transparent)
  PosX=w
  PosY=4
  
  
  String = "д 2 (ав. Конденсатор)"
  Width=TextWidth(String)
  PosX-Width
  
  DrawText(PosX, PosY, String, gParam\Color\Sensor2)
  PosX-20
  Box(PosX, PosY+TextHeight(String)/2, 16, 2, gParam\Color\Sensor2)
  
  PosX-10
  String = "д 1 (ав. ТСА)"
  Width=TextWidth(String)
  PosX-Width
  
  DrawText(PosX, PosY, String, gParam\Color\Sensor1)
  PosX-20
  Box(PosX, PosY+TextHeight(String)/2, 16, 2, gParam\Color\Sensor1)
  
  
  
  PosX=w
  PosY=10+TextHeight(String)
  
  
  String = "д6"
  Width=TextWidth(String)
  PosX-Width
  
  DrawText(PosX, PosY, String, gParam\Color\Sensor6)
  PosX-20
  Box(PosX, PosY+TextHeight(String)/2, 16, 2, gParam\Color\Sensor6)
  
  PosX-10
  String = "д 5 (Брага)"
  Width=TextWidth(String)
  PosX-Width
  
  DrawText(PosX, PosY, String, gParam\Color\Sensor5)
  PosX-20
  Box(PosX, PosY+TextHeight(String)/2, 16, 2, gParam\Color\Sensor5)
  
  PosX-10
  String = "д 4 (Куб)"
  Width=TextWidth(String)
  PosX-Width
  
  DrawText(PosX, PosY, String, gParam\Color\Sensor4)
  PosX-20
  Box(PosX, PosY+TextHeight(String)/2, 16, 2, gParam\Color\Sensor4)
  
  PosX-10
  String = "д 3 (Колонна)"
  Width=TextWidth(String)
  PosX-Width
  
  DrawText(PosX, PosY, String, gParam\Color\Sensor3)
  PosX-20
  Box(PosX, PosY+TextHeight(String)/2, 16, 2, gParam\Color\Sensor3)
  
  
  s=ListSize(gList())
  If DrawDate And s>2
    x=gParam\ScrollPos
    If x<0
      x=0
    ElseIf x>=s
      x=s-1
    EndIf
    If SelectElement(gList(), x)
      PosX=4
      PosY=4
      String = FormatDate("%dd/%mm/%yyyy  %hh:%ii:%ss", gList()\Date)
      Width=TextWidth(String)
      DrawText(PosX, PosY, String, gParam\Color\Text)
      x=gParam\VisibleTime+gParam\ScrollPos-1
      If x<0
        x=0
      ElseIf x>=s
        x=s-1
      EndIf
      If SelectElement(gList(), x)
        PosY=8+TextHeight(String)
        String = FormatDate("%dd/%mm/%yyyy  %hh:%ii:%ss", gList()\Date)
        DrawText(PosX, PosY, String, gParam\Color\Text)
      EndIf
      
    EndIf
  EndIf
  
EndProcedure

Procedure DrawGrid(*Grid.RECT) ;Сетка.
  Protected Pos.f, Steps.f
  Protected String.s, i, Temp

; Горизонтальные линии.
  
  DrawingMode(#PB_2DDrawing_Outlined)
  Box(*Grid\left-1, *Grid\top-1, *Grid\right-*Grid\left+2, *Grid\bottom-*Grid\top+2, gParam\Color\Grid)
  Box(*Grid\left, *Grid\top, *Grid\right-*Grid\left, *Grid\bottom-*Grid\top, gParam\Color\Grid)
  
  DrawingMode(#PB_2DDrawing_Transparent)
  
  Pos=*Grid\top
  Steps=(*Grid\bottom-*Grid\top)/18
  i=100
  Repeat
    If i>10
      Line(*Grid\left, Pos, *Grid\right-*Grid\left, 1, gParam\Color\Grid)
    EndIf
    
    String=Str(i)
    
    If i<100
      Temp=TextHeight(String)
      If i>10
        Temp/2
      Else
        Temp-4
      EndIf
    Else
      Temp=4
    EndIf
    DrawText(*Grid\left-TextWidth(String)-4, Pos - Temp,
             String, gParam\Color\Text)
    DrawText(*Grid\right+20-TextWidth(String), Pos - Temp,
             String, gParam\Color\Text)
    i-5
    If Pos>*Grid\bottom
      Break
    EndIf
    Pos+Steps
  ForEver
    
EndProcedure


Procedure GetCraphRect(w, h, *s.RECT) ; Координаты графика.
  Protected Result = #False
  
  If *s And w>0 And h>0
    *s\left=27
    *s\top=44
    *s\right=w-24
    *s\bottom=h-20
    Result = #True
  EndIf
  
  ProcedureReturn Result
EndProcedure

Procedure CreateMoreDirectory(Path.s) ; Создание множества папок по абсолютному пути.
  Protected Result, CurrentPath.s
  Protected Count, i, Dir.s
  
  Result=#True
  CurrentPath=""
  If Path<>""
    Dir=GetPathPart(Path)
    If FileSize(Dir)<>-2
      ReplaceString(Path, "/", "\", #PB_String_InPlace)
      Count=CountString(Path, "\")
      If Count>=2
        CurrentPath=StringField(Path, 1, "\")+"\"
        For i=2 To Count
          Dir = StringField(Path, i, "\")
          If Dir<>""
            CurrentPath+Dir+"\"
            If FileSize(CurrentPath)<>-2
              If CreateDirectory(CurrentPath)=0
                Result=#False
                Break
              EndIf
            EndIf
          EndIf
        Next i
      EndIf
    EndIf
  Else
    Result=#False
  EndIf
  
  ProcedureReturn Result
EndProcedure

Procedure SaveToFile(v) ; Сохранение графика в файл в виде картинки.
  Shared ThreadID
  Protected sTemp.Param, Im, w, h, Grid.RECT
  Protected Date, Path.s, File.s, i, Size
  
  CopyStructure(@gParam, @sTemp, Param)
  
  Date=Date()
  Path=GetPathPart(ProgramFilename())+"График\"+FormatDate("%dd_%mm_%yyyy__%hh_%ii_%ss", Date)+"\"
  Size=ListSize(gList())
  
  If CreateMoreDirectory(Path)
    gParam\LogFunct("Начато сохранение графика", #Log_Info)
    Repeat
      
      w=Size*#PointSec+52
      If w>#ImgMaxWidth
        w=#ImgMaxWidth
      EndIf
      
      gParam\VisibleTime=Round(w/#PointSec, #PB_Round_Down)
      File=Path+Str(Round(Size/gParam\VisibleTime, #PB_Round_Down)+1)+".png"
      Size-gParam\VisibleTime
      gParam\ScrollPos=Size
      
      If w<450 ; Минимальная шириа картинки.
        w=450
      EndIf
            
      Im=CreateImage(#PB_Any, w, GadgetHeight(gParam\GraphGadget), 24, gParam\Color\Back)
      If Im
        
        If StartDrawing(ImageOutput(Im))
          w=OutputWidth() ; Ширина
          h=OutputHeight(); и высота области рисования.
          DrawingFont(gParam\FontID)
          
          
          GetCraphRect(w, h, @Grid)
          
          DrawTop(Grid\right, Grid\top-2, #True)    ; Рисование содержимого выше графика.
          DrawGrid(@Grid)                           ; Сетка.
          DrawInfo(@Grid)                           ; Отрисовка графика.
          StopDrawing()
          
          If SaveImage(Im, File, #PB_ImagePlugin_PNG)
            gParam\LogFunct("График успешно сохранен в файле "+GetFilePart(File), #Log_Info)
          Else
            gParam\LogFunct("Не удалось сохранить график в файле "+GetFilePart(File), #Log_Error)
          EndIf
          
        Else
          gParam\LogFunct("Ошибка функции StartDrawing при сохранении графика", #Log_Warning)
        EndIf
        
        FreeImage(Im)
      Else
        gParam\LogFunct("Не удалось создать изображения для сохранения графика", #Log_Warning)
        Break
      EndIf
      
    Until w<30000 Or Size<=60
    gParam\LogFunct("Сохранение графика завершено", #Log_Info)
  Else
    gParam\LogFunct("Не удалось создать папку для сохранения графика", #Log_Warning)
  EndIf
  
  CopyStructure(@sTemp, @gParam, Param)
  ClearList(gList())
  
  ThreadID=0
EndProcedure

;-Public procedure.

Procedure GraphInit(*Graph.GraphInit)
  Shared ThreadID
  Protected Result=#False, w, h, r.RECT
  
  ThreadID=0
          
  If *Graph
    If IsGadget(gParam\GraphGadget) And IsGadget(gParam\ScrollGadget)
      gParam\LogFunct = *Graph\LogFunct
      gParam\GraphGadget = *Graph\GraphGadget
      gParam\ScrollGadget = *Graph\ScrollGadget
      gParam\FontID=LoadFont(*Graph\Font, *Graph\FontName, *Graph\FontSize)
        
        If StartDrawing(CanvasOutput(gParam\GraphGadget))
          w=OutputWidth() ; Ширина
          h=OutputHeight(); и высота области рисования.
          StopDrawing()
          
          GetCraphRect(w, h, @r)
          
          r\top+1
          r\right-2
          r\bottom-2
          
          gParam\VisibleTime=Round((r\right-r\left)/#PointSec, #PB_Round_Up)
          gParam\MinTermo = 100
          gParam\MaxTermo = 1000
          gParam\PointY = (gParam\MaxTermo-gParam\MinTermo)/(r\bottom-r\top)
          
          SetGadgetAttribute(gParam\ScrollGadget, #PB_ScrollBar_Minimum, 0)
          SetGadgetAttribute(gParam\ScrollGadget, #PB_ScrollBar_Maximum, 0)
          SetGadgetAttribute(gParam\ScrollGadget, #PB_ScrollBar_PageLength, gParam\VisibleTime)
          SetGadgetState(gParam\ScrollGadget, 0)
          
          Result=#True
      
        EndIf
      
    EndIf
  EndIf
  
  ProcedureReturn Result
EndProcedure

Procedure GraphSetting(*s.GraphSetting)
  Shared ThreadID
  
  If ThreadID
    ProcedureReturn 0 ; Во время сохранения графика, работа этой процедуры невозможна.
  EndIf
  
  If *s
    CopyStructure(*s, gParam, GraphSetting)
  EndIf
EndProcedure


Procedure AddData(*Data.GraphData, Date) ; Добавление данных для графика.
  Shared ThreadID
  Static OldDay
  Protected Temp, x
  
  If ThreadID
    ProcedureReturn 0 ; Во время сохранения графика, работа этой процедуры невозможна.
  EndIf
    
  If Day(Date)<>OldDay
    If gParam\NewDay And ListSize(gList())>=60
      ThreadID=CreateThread(@SaveToFile(), 0) ; Сохранение графика в файл в виде картинки.
      If ThreadID=0
        ClearList(gList())
        gParam\LogFunct("Не удалось создать параллельный поток для сохранения графика", #Log_Error)
      EndIf
    Else
      ClearList(gList())
    EndIf
    
    OldDay = Day(Date)
    SetGadgetAttribute(gParam\ScrollGadget, #PB_ScrollBar_Minimum, 0)
    SetGadgetAttribute(gParam\ScrollGadget, #PB_ScrollBar_Maximum, 0)
    SetGadgetAttribute(gParam\ScrollGadget, #PB_ScrollBar_PageLength, gParam\VisibleTime)
    SetGadgetState(gParam\ScrollGadget, 0)
    gParam\ScrollPos=0
  EndIf
  
  If ThreadID=0
    
    If *Data
      LastElement(gList())
      If AddElement(gList())
        CopyStructure(*Data, @gList()\Sensors, GraphData)
        gList()\Date = Date ; Дата добавления.
      EndIf
      SetGadgetAttribute(gParam\ScrollGadget, #PB_ScrollBar_Maximum, ListSize(gList())-1)
      SetGadgetAttribute(gParam\ScrollGadget, #PB_ScrollBar_PageLength, gParam\VisibleTime)
      
      If gParam\AutoScroll; Автопрокрутка графика.
        gParam\ScrollPos=ListSize(gList())-gParam\VisibleTime
        If gParam\ScrollPos<0
          gParam\ScrollPos=0
        EndIf
        SetGadgetState(gParam\ScrollGadget, gParam\ScrollPos)
      EndIf
      
    EndIf
    
    
    x=ListSize(gList())
    If x>200000
      If FirstElement(gList())
        Repeat
          If DeleteElement(gList(), 1)=0
            Break
          EndIf
          x-1
        Until x<=200000
      EndIf
    EndIf
    
  EndIf
EndProcedure



Procedure GraphDraw()
  Shared ThreadID
  Protected w, h, Grid.RECT
  Protected String.s
  
  If ThreadID
    ProcedureReturn 0 ; Во время сохранения графика, работа этой процедуры невозможна.
  EndIf
  
  If StartDrawing(CanvasOutput(gParam\GraphGadget))
    w=OutputWidth() ; Ширина
    h=OutputHeight(); и высота области рисования.
    DrawingFont(gParam\FontID)
    
    GetCraphRect(w, h, @Grid)
    
    Box(0, 0, w, h, gParam\Color\Back) ; Очитска области рисования.
    DrawTop(Grid\right, Grid\top-2)    ; Рисование содержимого выше графика.
    DrawGrid(@Grid)                    ; Сетка.
    DrawInfo(@Grid, #True)             ; Отрисовка графика.
    StopDrawing()
  EndIf
    
EndProcedure

Procedure Scroll_CB()
  Shared ThreadID
  
  If ThreadID
    ProcedureReturn 0 ; Во время сохранения графика, работа этой процедуры невозможна.
  EndIf
  
  gParam\ScrollPos = GetGadgetState(gParam\ScrollGadget)
  GraphDraw()
EndProcedure

EndModule

; IDE Options = PureBasic 5.31 (Windows - x86)
; Folding = ---
; EnableUnicode
; EnableXP